@@ -7,7 +7,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2019 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2020 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -382,8 +382,12 @@ class Query {
       auto attr_name = b_it.first;
       auto size_pair = b_it.second;
       auto var =
-          (attr_name != TILEDB_COORDS &&
-           schema_.attribute(attr_name).cell_val_num() == TILEDB_VAR_NUM);
+          ((attr_name != TILEDB_COORDS) &&
+           ((schema_.has_attribute(attr_name) &&
+             schema_.attribute(attr_name).cell_val_num() == TILEDB_VAR_NUM) ||
+            (schema_.domain().has_dimension(attr_name) &&
+             schema_.domain().dimension(attr_name).cell_val_num() ==
+                 TILEDB_VAR_NUM)));
       auto element_size = element_sizes_.find(attr_name)->second;
       elements[attr_name] = (var) ? std::pair<uint64_t, uint64_t>(
                                         size_pair.first / sizeof(uint64_t),
@@ -418,7 +422,7 @@ class Query {
    */
   template <class T>
   Query& add_range(uint32_t dim_idx, T start, T end, T stride = 0) {
-    impl::type_check<T>(schema_.domain().type());
+    impl::type_check<T>(schema_.domain().dimension(dim_idx).type());
     auto& ctx = ctx_.get();
     ctx.handle_error(tiledb_query_add_range(
         ctx.ptr().get(),
@@ -430,6 +434,42 @@ class Query {
     return *this;
   }
 
+  /**
+   * Adds a 1D string range along a subarray dimension, in the form
+   * (start, end). Applicable only to variable-sized dimensions
+   *
+   * **Example:**
+   *
+   * @code{.cpp}
+   * // Set a 1D range on dimension 0, assuming the domain type is int64.
+   * int64_t start = 10;
+   * int64_t end = 20;
+   * // Stride is optional
+   * subarray.add_range(0, start, end);
+   * @endcode
+   *
+   * @tparam T The dimension datatype
+   * @param dim_idx The index of the dimension to add the range to.
+   * @param start The range start to add.
+   * @param end The range end to add.
+   * @param stride The range stride to add.
+   * @return Reference to this Query
+   */
+  Query& add_range(
+      uint32_t dim_idx, const std::string& start, const std::string& end) {
+    impl::type_check<char>(schema_.domain().dimension(dim_idx).type());
+    auto& ctx = ctx_.get();
+    ctx.handle_error(tiledb_query_add_range_var(
+        ctx.ptr().get(),
+        query_.get(),
+        dim_idx,
+        start.c_str(),
+        start.size(),
+        end.c_str(),
+        end.size()));
+    return *this;
+  }
+
   /**
    * Retrieves the number of ranges for a given dimension.
    *
@@ -471,7 +511,7 @@ class Query {
    */
   template <class T>
   std::array<T, 3> range(unsigned dim_idx, uint64_t range_idx) {
-    impl::type_check<T>(schema_.domain().type());
+    impl::type_check<T>(schema_.domain().dimension(dim_idx).type());
     auto& ctx = ctx_.get();
     const void *start, *end, *stride;
     ctx.handle_error(tiledb_query_get_range(
@@ -488,6 +528,44 @@ class Query {
     return ret;
   }
 
+  /**
+   * Retrieves a range for a given variable length string dimension and range
+   * id.
+   *
+   * **Example:**
+   *
+   * @code{.cpp}
+   * unsigned dim_idx = 0;
+   * unsigned range_idx = 0;
+   * std::array<std::string, 2> range = query.range(dim_idx, range_idx);
+   * @endcode
+   *
+   * @param dim_idx The dimension index.
+   * @param range_idx The range index.
+   * @return A pair of the form (start, end).
+   */
+  std::array<std::string, 2> range(unsigned dim_idx, uint64_t range_idx) {
+    impl::type_check<char>(schema_.domain().dimension(dim_idx).type());
+    auto& ctx = ctx_.get();
+    uint64_t start_size, end_size;
+    ctx.handle_error(tiledb_query_get_range_var_size(
+        ctx.ptr().get(),
+        query_.get(),
+        dim_idx,
+        range_idx,
+        &start_size,
+        &end_size));
+
+    std::string start;
+    start.resize(start_size);
+    std::string end;
+    end.resize(end_size);
+    ctx.handle_error(tiledb_query_get_range_var(
+        ctx.ptr().get(), query_.get(), dim_idx, range_idx, &start[0], &end[0]));
+    std::array<std::string, 2> ret = {{std::move(start), std::move(end)}};
+    return ret;
+  }
+
   /**
    * Retrieves the estimated result size for a fixed-size attribute.
    *
@@ -678,6 +756,10 @@ class Query {
   /**
    * Set the coordinate buffer.
    *
+   * The coordinate buffer has been deprecated. Set the coordinates for
+   * each individual dimension with the `set_buffer` API. Consult the current
+   * documentation for more information.
+   *
    * **Example:**
    * @code{.cpp}
    * tiledb::Context ctx;
@@ -695,12 +777,17 @@ class Query {
    * @param size The number of elements in the coordinate array buffer
    * **/
   template <typename T>
-  Query& set_coordinates(T* buf, uint64_t size) {
+  TILEDB_DEPRECATED Query& set_coordinates(T* buf, uint64_t size) {
     impl::type_check<T>(schema_.domain().type());
     return set_buffer(TILEDB_COORDS, buf, size);
   }
 
-  /** Set the coordinate buffer for unordered queries
+  /**
+   * Set the coordinate buffer for unordered queries.
+   *
+   * The coordinate buffer has been deprecated. Set the coordinates for
+   * each individual dimension with the `set_buffer` API. Consult the current
+   * documentation for more information.
    *
    * **Example:**
    * @code{.cpp}
@@ -716,12 +803,12 @@ class Query {
    * @param buf Coordinate vector
    * **/
   template <typename Vec>
-  Query& set_coordinates(Vec& buf) {
+  TILEDB_DEPRECATED Query& set_coordinates(Vec& buf) {
     return set_coordinates(buf.data(), buf.size());
   }
 
   /**
-   * Sets a buffer for a fixed-sized attribute.
+   * Sets a buffer for a fixed-sized attribute/dimension.
    *
    * **Example:**
    * @code{.cpp}
@@ -734,20 +821,33 @@ class Query {
    *
    * @note set_buffer(std::string, std::vector) is preferred as it is safer.
    *
-   * @tparam T Attribute value type
-   * @param attr Attribute name
-   * @param buff Buffer array pointer with elements of the attribute type.
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
+   * @param buff Buffer array pointer with elements of the
+   *     attribute/dimension type.
    * @param nelements Number of array elements
    **/
   template <typename T>
-  Query& set_buffer(const std::string& attr, T* buff, uint64_t nelements) {
-    if (attr != TILEDB_COORDS)
-      impl::type_check<T>(schema_.attribute(attr).type());
-    return set_buffer(attr, buff, nelements, sizeof(T));
+  Query& set_buffer(const std::string& name, T* buff, uint64_t nelements) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (name != TILEDB_COORDS && !is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<T>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<T>(schema_.domain().dimension(name).type());
+    else if (name == TILEDB_COORDS)
+      impl::type_check<T>(schema_.domain().type());
+
+    return set_buffer(name, buff, nelements, sizeof(T));
   }
 
   /**
-   * Sets a buffer for a fixed-sized attribute.
+   * Sets a buffer for a fixed-sized attribute/dimension.
    *
    * **Example:**
    * @code{.cpp}
@@ -758,37 +858,50 @@ class Query {
    * query.set_buffer("a1", data_a1);
    * @endcode
    *
-   * @tparam T Attribute value type
-   * @param attr Attribute name
-   * @param buf Buffer vector with elements of the attribute type.
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
+   * @param buf Buffer vector with elements of the attribute/dimension type.
    **/
   template <typename T>
-  Query& set_buffer(const std::string& attr, std::vector<T>& buf) {
-    return set_buffer(attr, buf.data(), buf.size(), sizeof(T));
+  Query& set_buffer(const std::string& name, std::vector<T>& buf) {
+    return set_buffer(name, buf.data(), buf.size(), sizeof(T));
   }
 
   /**
-   * Sets a buffer for a fixed-sized attribute.
+   * Sets a buffer for a fixed-sized attribute/dimension.
    *
    * @note This unsafe version does not perform type checking; the given buffer
    * is assumed to be the correct type, and the size of an element in the given
    * buffer is assumed to be the size of the datatype of the attribute.
    *
-   * @param attr Attribute name
+   * @param nam Attribute/Dimension name
    * @param buff Buffer array pointer with elements of the attribute type.
    * @param nelements Number of array elements in buffer
    **/
-  Query& set_buffer(const std::string& attr, void* buff, uint64_t nelements) {
+  Query& set_buffer(const std::string& name, void* buff, uint64_t nelements) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (name != TILEDB_COORDS && !is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+
     // Compute element size (in bytes).
-    size_t element_size =
-        attr == TILEDB_COORDS ?
-            tiledb_datatype_size(schema_.domain().type()) :
-            tiledb_datatype_size(schema_.attribute(attr).type());
-    return set_buffer(attr, buff, nelements, element_size);
+    size_t element_size = 0;
+    if (name == TILEDB_COORDS)
+      element_size = tiledb_datatype_size(schema_.domain().type());
+    else if (is_attr)
+      element_size = tiledb_datatype_size(schema_.attribute(name).type());
+    else if (is_dim)
+      element_size =
+          tiledb_datatype_size(schema_.domain().dimension(name).type());
+
+    return set_buffer(name, buff, nelements, element_size);
   }
 
   /**
-   * Sets a buffer for a variable-sized attribute.
+   * Sets a buffer for a variable-sized attribute/dimension.
    *
    * **Example:**
    *
@@ -804,8 +917,8 @@ class Query {
    * @note set_buffer(std::string, std::vector, std::vector) is preferred as it
    * is safer.
    *
-   * @tparam T Attribute value type
-   * @param attr Attribute name
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
    * @param offsets Offsets array pointer where a new element begins in the data
    *        buffer.
    * @param offsets_nelements Number of elements in offsets buffer.
@@ -815,24 +928,35 @@ class Query {
    **/
   template <typename T>
   Query& set_buffer(
-      const std::string& attr,
+      const std::string& name,
       uint64_t* offsets,
       uint64_t offset_nelements,
       T* data,
       uint64_t data_nelements) {
-    impl::type_check<T>(schema_.attribute(attr).type());
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<T>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<T>(schema_.domain().dimension(name).type());
+
     return set_buffer(
-        attr, offsets, offset_nelements, data, data_nelements, sizeof(T));
+        name, offsets, offset_nelements, data, data_nelements, sizeof(T));
   }
 
   /**
-   * Sets a buffer for a variable-sized attribute.
+   * Sets a buffer for a variable-sized attribute/dimension.
    *
    * @note This unsafe version does not perform type checking; the given buffer
    * is assumed to be the correct type, and the size of an element in the given
    * buffer is assumed to be the size of the datatype of the attribute.
    *
-   * @param attr Attribute name
+   * @param name Attribute/Dimension name
    * @param offsets Offsets array pointer where a new element begins in the data
    *        buffer.
    * @param offsets_nelements Number of elements in offsets buffer.
@@ -840,19 +964,30 @@ class Query {
    * @param data_nelements Number of array elements in data buffer.
    **/
   Query& set_buffer(
-      const std::string& attr,
+      const std::string& name,
       uint64_t* offsets,
       uint64_t offset_nelements,
       void* data,
       uint64_t data_nelements) {
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+
     // Compute element size (in bytes).
-    size_t element_size = tiledb_datatype_size(schema_.attribute(attr).type());
+    auto type = is_attr ? schema_.attribute(name).type() :
+                          schema_.domain().dimension(name).type();
+    size_t element_size = tiledb_datatype_size(type);
+
     return set_buffer(
-        attr, offsets, offset_nelements, data, data_nelements, element_size);
+        name, offsets, offset_nelements, data, data_nelements, element_size);
   }
 
   /**
-   * Sets a buffer for a variable-sized attribute.
+   * Sets a buffer for a variable-sized attribute/dimension.
    *
    * **Example:**
    * @code{.cpp}
@@ -864,8 +999,8 @@ class Query {
    * query.set_buffer("a1", offsets_a1, data_a1);
    * @endcode
    *
-   * @tparam T Attribute value type
-   * @param attr Attribute name
+   * @tparam T Attribute/Dimension value type
+   * @param name Attribute/Dimension name
    * @param offsets Offsets where a new element begins in the data buffer.
    * @param data Buffer vector with elements of the attribute type.
    *        For variable sized attributes, the buffer should be flattened. E.x.
@@ -874,43 +1009,76 @@ class Query {
    **/
   template <typename T>
   Query& set_buffer(
-      const std::string& attr,
+      const std::string& name,
       std::vector<uint64_t>& offsets,
       std::vector<T>& data) {
-    impl::type_check<T>(schema_.attribute(attr).type());
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<T>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<T>(schema_.domain().dimension(name).type());
+
     return set_buffer(
-        attr, offsets.data(), offsets.size(), &data[0], data.size(), sizeof(T));
+        name, offsets.data(), offsets.size(), &data[0], data.size(), sizeof(T));
   }
 
   /**
-   * Sets a buffer for a variable-sized attribute.
+   * Sets a buffer for a variable-sized attribute/dimension.
    *
-   * @tparam T Attribute value type
-   * @param attr Attribute name
-   * @param buf pair of offset, data buffers
+   * @tparam T Attribute/Dimension value type
+   * @param attr Attribute/Dimension name
+   * @param buf Pair of offset, data buffers
    **/
   template <typename T>
   Query& set_buffer(
-      const std::string& attr,
+      const std::string& name,
       std::pair<std::vector<uint64_t>, std::vector<T>>& buf) {
-    impl::type_check<T>(schema_.attribute(attr).type());
-    return set_buffer(attr, buf.first, buf.second);
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<T>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<T>(schema_.domain().dimension(name).type());
+
+    return set_buffer(name, buf.first, buf.second);
   }
 
   /**
-   * Sets a buffer for a string-typed variable-sized attribute.
+   * Sets a buffer for a string-typed variable-sized attribute/dimension.
    *
-   * @param attr Attribute name
+   * @param name Attribute/Dimension name
    * @param offsets Offsets where a new element begins in the data buffer.
    * @param data Pre-allocated string buffer.
    **/
   Query& set_buffer(
-      const std::string& attr,
+      const std::string& name,
       std::vector<uint64_t>& offsets,
       std::string& data) {
-    impl::type_check<char>(schema_.attribute(attr).type());
+    // Checks
+    auto is_attr = schema_.has_attribute(name);
+    auto is_dim = schema_.domain().has_dimension(name);
+    if (!is_attr && !is_dim)
+      throw TileDBError(
+          std::string("Cannot set buffer; Attribute/Dimension '") + name +
+          "' does not exist");
+    else if (is_attr)
+      impl::type_check<char>(schema_.attribute(name).type());
+    else if (is_dim)
+      impl::type_check<char>(schema_.domain().dimension(name).type());
+
     return set_buffer(
-        attr,
+        name,
         offsets.data(),
         offsets.size(),
         &data[0],
